"""
ProjectFile
"""
import yaml
import datetime
import os
from mlops.utils.logger import logger


class ProjectFile:

    def __init__(self, config, path='.', projectfile_name='MLproject', use_localhost=False, clean_projectfile=True):
        self.config = config
        self.projectfile_name = projectfile_name
        self.project_path = os.path.join(path, projectfile_name)

        # Clean projectfile if it exists
        if clean_projectfile and os.path.exists(self.project_path):
            logger.debug('Removing existing project file: {0}'.format(self.project_path))
            os.remove(os.path.join(path, self.projectfile_name))

        if use_localhost:
            volume_path = config['project']['LOCAL_VOLUME_MOUNT']
        else:
            volume_path = config['project']['VOLUME_MOUNT']

        self.project_dict = {'name': config['project']['NAME'].lower(),
                             'docker_env': {'volumes': [volume_path],
                                            'image': config['project']['NAME'].lower()}}

        self._parse_entry_points()

    def _parse_entry_points(self):
        entry_points_dict = {}
        for key in self.config['entry_points']:
            entry_points_dict[key] = {'command': self.config['entry_points'][key]}
        self.project_dict['entry_points'] = entry_points_dict

    def generate_yaml(self):
        logger.info('Writing project file: {0}'.format(self.project_path))
        with open(self.project_path, 'w+') as f:
            f.write('# {0}: This ProjectFile was automatically generated by MLOps\n\n'.format(datetime.datetime.now()))
            yaml.dump(self.project_dict, f, default_flow_style=False, sort_keys=False)



from mlops.Experiment import Experiment


class TestExperiment:

    def setup(self):
        # currently only testing localhost code
        use_localhost = True
        self.experiment = Experiment('tests/data/test_config.cfg', use_localhost=use_localhost)

    def test_log_artifact(self):
        """Test to log artifact and then check it is available on MINIO"""
        pass

    def test_log_artifact(self):
        """Test to log artifact and then check it is available on MINIO"""
        pass



import logging
import colorlog
import sys


def configure_logger():

    # make log formatters
    stream_formatter = colorlog.ColoredFormatter('%(log_color)s%(asctime)-15s %(levelname).1s '
                                                 '[%(filename)s:%(funcName)s:%(lineno)d] %(message)s',
                                                 '%Y-%m-%d %H:%M:%S')

    file_formatter = logging.Formatter('%(asctime)-15s %(levelname).1s [%(filename)s:%(funcName)s:%(lineno)d]'
                                       ' %(message)s',
                                       '%Y-%m-%d %H:%M:%S')

    log_file = 'MLOps_logger.log'
    stream_handler = logging.StreamHandler(sys.stdout)
    file_handler = logging.FileHandler(log_file)

    # set formatters
    stream_handler.setFormatter(stream_formatter)
    file_handler.setFormatter(file_formatter)

    # add handlers
    logger.addHandler(stream_handler)
    logger.addHandler(file_handler)

    # set logging level
    file_handler.setLevel(logging.DEBUG)
    stream_handler.setLevel(logging.INFO)

    logger.setLevel(logging.DEBUG)


logger = logging.getLogger(__name__)
configure_logger()



from mlops.Experiment import Experiment
import configparser
from minio import Minio
import os
import docker
import pytest


class TestExperiment:

    def setup(self):
        # currently only testing localhost code
        self.experiment = Experiment('tests/data/test_config.cfg', project_path='tests/data', use_localhost=True)

    def test_check_dirty(self):
        """
        Test that the local and remote experiments are the same
        """
        # Need to set project_path at level of git directory for this test.
        self.experiment = Experiment('tests/data/test_config.cfg', project_path='.', use_localhost=True)
        assert not self.experiment.check_dirty()

    def test_check_environment_variables(self):
        test_var = os.environ['AWS_ACCESS_KEY_ID']
        del os.environ['AWS_ACCESS_KEY_ID']
        with pytest.raises(Exception) as e:
            self.experiment.check_environment_variables()
        # reset env var
        os.environ['AWS_ACCESS_KEY_ID'] = test_var

    def test_config_setup(self):
        self.experiment.config_setup()
        assert self.experiment.experiment_name == 'test_project'
        assert self.experiment.artifact_path == 's3://mlflow'

    def test_env_setup(self):
        self.experiment.env_setup()
        assert os.getenv('MLFLOW_TRACKING_URI') == self.experiment.config['server']['LOCAL_REMOTE_SERVER_URI']
        assert os.getenv('MLFLOW_S3_ENDPOINT_URL') == self.experiment.config['server']['LOCAL_MLFLOW_S3_ENDPOINT_URL']

    def test_read_config(self):
        # Create config file and assert identical
        self.test_config = configparser.ConfigParser()
        self.test_config.read(self.experiment.config_path)
        assert self.experiment.config == self.test_config

    def test_init_experiment(self):
        self.experiment.init_experiment()
        assert self.experiment.experiment_id == '1'
        self.experiment.experiment_name = 'test_project_init_experiment'
        self.experiment.init_experiment()
        assert self.experiment.experiment_id == '2'

    def test_print_experiment_info(self, caplog):
        # Check correct information is printed to console
        self.experiment.print_experiment_info()  # Call function.
        assert 'Name: test_project' in caplog.text
        assert 'Artifact Location: s3://mlflow' in caplog.text

    def test_configure_minio(self):
        # check mlflow bucket is created
        self.experiment.configure_minio()
        client = Minio(self.experiment.uri_formatted,
                       self.experiment.minio_cred['user'],
                       self.experiment.minio_cred['password'],
                       secure=False)
        assert 'mlflow' in (bucket.name for bucket in client.list_buckets())

    def test_build_project_file(self):
        if os.path.exists('MLproject'):
            os.remove('MLproject')
        self.experiment.build_project_file()
        assert os.path.exists(os.path.join(self.experiment.project_path, 'MLproject'))

    def test_build_experiment_image(self):
        client = docker.from_env()
        images_1 = [img['RepoTags'][0] for img in client.api.images()]
        # assert self.experiment.experiment_name + ':latest' not in images_1
        self.experiment.build_experiment_image()
        images_2 = [img['RepoTags'][0] for img in client.api.images()]
        assert self.experiment.experiment_name + ':latest' in images_2

    def test_run(self, capsys):
        os.getcwd()
        self.experiment.build_project_file()
        self.experiment.run(path='tests/data/')
        captured = capsys.readouterr()
        assert 'succeeded' in captured.err



import pytest
import xnat
from mlops.data.transforms.LoadImageXNATd import LoadImageXNATd
from mlops.data.tools.tools import xnat_build_dataset
from monai.transforms import Compose, ToTensord
from torch.utils.data import DataLoader
from monai.data import CacheDataset
from xnat.mixin import ImageScanData, SubjectData
import requests
from xnat.exceptions import XNATUploadError
from requests.auth import HTTPBasicAuth


@pytest.mark.skip(reason="unable to test XNAT features on github actions currently. Not been able to create a connection to the test XNAT instance")
class TestLoadImageXNATd:

    def setup(self):
        self.test_batch_size = 1
        self.xnat_configuration = {'server': 'http://localhost:80',
                                   'user': 'admin',
                                   'password': 'admin',
                                   'project': 'TEST_MLOPS'}

        #  create test project and push test data
        with xnat.connect(server=self.xnat_configuration['server'],
                          user=self.xnat_configuration['user'],
                          password=self.xnat_configuration['password'], ) as session:

            xnat_projects_url = self.xnat_configuration['server'] + '/data/projects'

            # Set the name of the XML file.
            headers = {'Content-Type': 'text/xml'}

            # Open the XML file.
            with open('tests/data/xnat/test_project_setup.xml') as xml:
                r = requests.post(xnat_projects_url, data=xml, headers=headers,
                                  auth=HTTPBasicAuth(self.xnat_configuration['user'],
                                                     self.xnat_configuration['password']))

            # push data to new project
            try:
                session.services.import_('tests/data/test_dicoms.zip', project=self.xnat_configuration['project'],
                                         subject='1',
                                         experiment='MR_TEST_EXPERIMENT')
            except XNATUploadError as e:
                print(f'Test subject already exists {e}')

        self.test_data = xnat_build_dataset(self.xnat_configuration)

    def test_xnat_connection(self):
        try:
            r = requests.get(self.xnat_configuration['server'])
            r.raise_for_status()  # Raises a HTTPError if the status is 4xx, 5xxx
        except Exception as e:
            raise e
        else:
            pass  # Proceed to do stuff with `r`

    def test_create_dataloader_with_transform(self):

        def fetch_test_exp(subject_data: SubjectData = None) -> (ImageScanData, str):
            """
            Function that identifies and returns the required xnat ImageData object from a xnat SubjectData object
            along with the 'key' that it will be used to access it.
            """
            for exp in subject_data.experiments:
                if 'MR_TEST_EXPERIMENT' in subject_data.experiments[exp].label:
                    for scan in subject_data.experiments[exp].scans:
                        if 'SlicePosition' in subject_data.experiments[exp].scans[scan].series_description:
                            return subject_data.experiments[exp].scans[scan], 'test_image'

        # fetches are applied sequentially
        actions = [fetch_test_exp]

        self.train_transforms = Compose(
            [
                LoadImageXNATd(keys=['subject_uri'], actions=actions, xnat_configuration=self.xnat_configuration, expected_filetype='.IMA'),
                ToTensord(keys=['test_image'])
            ]
        )

        self.dataset = CacheDataset(data=self.test_data, transform=self.train_transforms)
        from monai.data.utils import list_data_collate
        self.loader = DataLoader(self.dataset, batch_size=self.test_batch_size,
                                 shuffle=True, num_workers=0, collate_fn=list_data_collate)

        for i_batch, sample_batched in enumerate(self.loader):
            assert 'test_image' in sample_batched
            assert len(sample_batched['test_image']) == self.test_batch_size



import os

import xnat
from mlops.utils.logger import logger

def xnat_build_dataset(xnat_configuration, minimal=True):
    """
    Builds a dictionary that describes the XNAT project dataset using XNAT data hierarchy: project/subject/experiment/scan

    minimal: if True, only the minimal information is returned, i.e. the  subject ID and the SubjectData

    e.g.
    [
        {
        subjectid: '1',
        subject_uri: '<unique address to subject in xnat archive>'
        experiment_1: {
                        scan_1: {scan_object},
                        scan_2: {scan_object}
                       },
        experiment_2:  {
                        scan_1: {scan_object}
                      }
        }
    ]

    """
    with xnat.connect(server=xnat_configuration['server'],
                      user=xnat_configuration['user'],
                      password=xnat_configuration['password'], verify=bool(os.getenv('MLOPS_SSL'))) as session:

        logger.info(f"Collecting XNAT project: {xnat_configuration['project']}")
        project = session.projects[xnat_configuration["project"]]

        dataset = []
        for subject in project.subjects:

            data_series = {'subject_id': project.subjects[subject].label, 'subject_uri': project.subjects[subject].uri}

            if not minimal:
                for experiment in project.subjects[subject].experiments:
                    scan_dict = {}
                    for scan in project.subjects[subject].experiments[experiment].scans:
                        scan_dict['modality'] = project.subjects[subject].experiments[experiment].modality
                        scan_dict[project.subjects[subject].experiments[experiment].scans[scan].series_description] = project.subjects[subject].experiments[experiment].scans[scan]

                    data_series[project.subjects[subject].experiments[experiment].label] = scan_dict
            logger.debug(f"Adding data series to dataset: {data_series}")
            dataset.append(data_series)

    return dataset




import setuptools
# from mlops import __version__

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

version = "0.5.8"

setuptools.setup(
    name="csc-mlops",
    version=version,
    author="Laurence Jackson",
    author_email="laurence.jackson@gstt.nhs.uk",
    description="An MLOps framework for development of clinical applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GSTT-CSC/MLOps",
    project_urls={
        "Bug Tracker": "https://github.com/GSTT-CSC/MLOps/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.8",
)



import mlflow
import os
import sys
import configparser
import docker
from minio import Minio
from mlops.ProjectFile import ProjectFile
from mlops.utils.logger import logger
from git import Repo


class Experiment:

    def __init__(self, config_path='config.cfg', project_path='.', use_localhost=False, verbose=True):

        self.config = None
        self.artifact_path = None
        self.experiment_name = None
        self.experiment_id = None
        self.config_path = config_path
        self.use_localhost = use_localhost
        self.project_path = project_path
        self.verbose = verbose

        if 'pytest' not in sys.modules:
            self.check_dirty()

        self.check_environment_variables()
        self.config_setup()
        self.env_setup()
        self.build_project_file()
        self.init_experiment()

        if self.verbose:
            self.print_experiment_info()

    def check_dirty(self):
        logger.debug('Comparing to remote git repository')
        repo = Repo(self.project_path)
        head = repo.head.ref
        tracking = head.tracking_branch()
        local_commits_ahead_iter = head.commit.iter_items(repo, f'{tracking.path}..{head.path}')
        commits_ahead = sum(1 for _ in local_commits_ahead_iter)

        if repo.is_dirty():
            raise Exception('Repository is dirty. Please commit your changes before running the experiment')
        if commits_ahead > 0:
            raise Exception('Local repository ahead of remote. Please push changes before running the experiment')

        if not repo.is_dirty() and commits_ahead == 0:
            return False
        else:
            raise Exception('Please synchronise local and remote code versions before running the experiment')

    @staticmethod
    def check_environment_variables():
        required_env_variables = ['AWS_ACCESS_KEY_ID',
                                  'AWS_SECRET_ACCESS_KEY']

        for var in required_env_variables:
            if os.getenv(var) is None:
                raise Exception('{0} is a required environment variable: set with "export {0}=<value>"'.format(var))

    def config_setup(self):
        logger.info('reading config file: {0}'.format(self.config_path))
        self.read_config()
        self.artifact_path = self.config['server']['ARTIFACT_PATH']
        self.experiment_name = self.config['project']['NAME'].lower()

    def env_setup(self):

        if self.use_localhost:
            os.environ['MLFLOW_TRACKING_URI'] = self.config['server']['LOCAL_REMOTE_SERVER_URI']
            os.environ['MLFLOW_S3_ENDPOINT_URL'] = self.config['server']['LOCAL_MLFLOW_S3_ENDPOINT_URL']
        else:
            os.environ['MLFLOW_TRACKING_URI'] = self.config['server']['REMOTE_SERVER_URI']
            os.environ['MLFLOW_S3_ENDPOINT_URL'] = self.config['server']['MLFLOW_S3_ENDPOINT_URL']

    def read_config(self):
        self.config = configparser.ConfigParser()
        self.config.read(self.config_path)

    def init_experiment(self):
        # logger.info('Creating experiment: name: {0} *** ID: {1}'.format(self.experiment_name, exp_id))
        experiment = mlflow.get_experiment_by_name(self.experiment_name)

        if experiment is None:
            exp_id = mlflow.create_experiment(self.experiment_name, artifact_location=self.artifact_path)
            logger.info('Creating experiment: name: {0} *** ID: {1}'.format(self.experiment_name, exp_id))
        else:
            exp_id = experiment.experiment_id
            logger.info('Logging to existing experiment: {0} *** ID: {1}'.format(self.experiment_name, exp_id))

        logger.info('Setting tracking URI to: {0} '.format(os.environ['MLFLOW_TRACKING_URI']))
        mlflow.set_tracking_uri(os.environ['MLFLOW_TRACKING_URI'])
        logger.info('Setting experiment to: {0} '.format(self.experiment_name))
        mlflow.set_experiment(self.experiment_name)
        self.configure_minio()
        self.experiment_id = exp_id

    def print_experiment_info(self):
        experiment = mlflow.get_experiment(self.experiment_id)
        logger.info("Name: {}".format(experiment.name))
        logger.info("Experiment_id: {}".format(experiment.experiment_id))
        logger.info("Artifact Location: {}".format(experiment.artifact_location))
        logger.info("Tags: {}".format(experiment.tags))
        logger.info("Lifecycle_stage: {}".format(experiment.lifecycle_stage))

    def configure_minio(self):
        if self.use_localhost:
            self.uri_formatted = self.config['server']['LOCAL_MLFLOW_S3_ENDPOINT_URL'].replace("http://", "")
        else:
            self.uri_formatted = self.config['server']['MLFLOW_S3_ENDPOINT_URL'].replace("http://", "")

        self.minio_cred = {'user': os.getenv('AWS_ACCESS_KEY_ID'),
                           'password': os.getenv('AWS_SECRET_ACCESS_KEY')}

        # todo: replace this with either a machine level IAM role or ~/.aws/credentials profile
        os.environ['MINIO_ROOT_USER'] = os.getenv('AWS_ACCESS_KEY_ID')
        os.environ['MINIO_ROOT_PASSWORD'] = os.getenv('AWS_SECRET_ACCESS_KEY')

        client = Minio(self.uri_formatted, self.minio_cred['user'], self.minio_cred['password'], secure=False)
        # if mlflow bucket does not exist, create it
        if 'mlflow' not in (bucket.name for bucket in client.list_buckets()):
            logger.info('Creating S3 bucket ''mlflow''')
            client.make_bucket("mlflow")

    def build_experiment_image(self, path: str = '.'):
        logger.info('Building experiment image ...')

        # Collect proxy settings
        build_args = {}
        if os.getenv('http_proxy') is not None or os.getenv('https_proxy') is not None:
            build_args = {'http_proxy': os.getenv('http_proxy'),
                          'https_proxy': os.getenv('https_proxy')}

        client = docker.from_env()
        logger.info('Running docker build with: {0}'.format({'path': self.project_path,
                                                              'tag': self.experiment_name,
                                                              'buildargs': build_args,
                                                              'rm': 'True'}))

        client.images.build(path=self.project_path,
                            tag=self.experiment_name,
                            buildargs=build_args,
                            rm=True)
        logger.info('Built: ' + self.experiment_name + ':latest')

    def build_project_file(self, path: str = '.'):
        logger.info('Building project file')
        projectfile = ProjectFile(self.config, path=self.project_path, use_localhost=self.use_localhost)
        projectfile.generate_yaml()

    def run(self, path: str = '.', remote: str = None, **kwargs):
        logger.info('Starting experiment ...')

        docker_args_default = {'network': "host",
                               'ipc': 'host',
                               'rm': ''
                               }

        if not self.use_localhost:
            gpu_params = {'gpus': 'all',
                          'runtime': 'nvidia'}
            logger.info('Adding docker args: {0}'.format(gpu_params))
            docker_args_default.update(gpu_params)

        # update docker_args_default with values passed by project
        if 'docker_args' in kwargs:
            docker_args_default.update(kwargs['docker_args'])
            kwargs['docker_args'] = docker_args_default

        # check image exists and build if not
        logger.info('Checking for existing image')
        client = docker.from_env()
        images = [str(img['RepoTags']) for img in client.api.images()]
        if all([(self.experiment_name + ':latest') not in item for item in images]):
            logger.info('No existing image found')
            self.build_experiment_image(path=self.project_path)
        else:
            logger.info('Found existing project image')

        artifact_uri = mlflow.get_artifact_uri()
        print("Artifact uri: {}".format(artifact_uri))

        mlflow.run(self.project_path,
                   experiment_id=self.experiment_id,
                   use_conda=False,
                   **kwargs)



"""
MONAI MapTransform for importing image data from XNAT
"""
import tempfile
import xnat
import os
import glob
from monai.transforms import MapTransform, LoadImage
from monai.config import KeysCollection
from mlops.utils.logger import logger


class LoadImageXNATd(MapTransform):
    """
    MapTransform for importing image data from XNAT
    """

    def __init__(self, keys: KeysCollection, actions=None, xnat_configuration=None, image_loader=LoadImage, expected_filetype='.dcm'):
        super().__init__(keys)
        self.image_loader = image_loader
        self.xnat_configuration = xnat_configuration
        self.actions = actions
        self.expected_filetype = expected_filetype

    def __call__(self, data):
        d = dict(data)
        for key in self.keys:
            if key in data:

                for action in self.actions:
                    """loops over actions in action list, if no action is triggered then raise a warning actions are 
                    functions that return any of projects, subjects, experiments, scans, or resources XNAT object 
                    along with a key to be used in the data dictionary"""

                    logger.debug(f"Running XNAT action: {action}")

                    with xnat.connect(server=self.xnat_configuration['server'],
                                      user=self.xnat_configuration['user'],
                                      password=self.xnat_configuration['password'], verify=bool(os.getenv('MLOPS_SSL'))) as session:

                        "connect session to subject uri"
                        subject_obj = session.create_object(d['subject_uri'])
                        "perform action on subject object"
                        xnat_obj, image_label = action(subject_obj)

                        with tempfile.TemporaryDirectory() as tmpdirname:
                            "download image from XNAT"
                            session_obj = session.create_object(xnat_obj.uri)
                            session_obj.download_dir(tmpdirname)

                            images_path = glob.glob(os.path.join(tmpdirname, '**/*' + self.expected_filetype), recursive=True)

                            "find unique directories in list of image paths"
                            image_dirs = list(set(os.path.dirname(image_path) for image_path in images_path))

                            if len(image_dirs) > 1:
                                raise ValueError(f'More than one image series found in {images_path}')

                            logger.info(f"Downloading image: {image_dirs[0]}")

                            image, meta = self.image_loader()(image_dirs[0])

                            d[image_label] = image
                            d[image_label + '_meta'] = meta

            return d
